;;;
;;; tokyocabinet
;;;

(define-module dbm.tokyocabinet
  (extend dbm)
  (export <tcbdb> <tchdb>

	  ;; low level interface
	  tcbdb-new tcbdb-ecode
	  tcbdb-open tcbdb-close
	  tcbdb-put tcbdb-get tcbdb-out2

	  tchdb-new tchdb-ecode
	  tchdb-open tchdb-close
	  tchdb-put2 tchdb-get2 tchdb-out2

	  ;; B+ tree specific API
	  tcbdbcur-new
	  tcbdbcur-first tcbdbcur-last
	  tcbdbcur-prev tcbdbcur-next

	  ;; Hash specific API
	  tchdb-iterinit
	  tchdb-iternext2 tchdb-iternext3

	  |BDBOREADER| |BDBOWRITER| |BDBOCREAT|
	  |BDBOTRUNC| |BDBONOLCK| |BDBOLCKNB|
	  
	  |HDBOREADER| |HDBOWRITER| |HDBOCREAT|
	  |HDBOTRUNC| |HDBONOLCK| |HDBOLCKNB|
	  
	  |TCESUCCESS| |TCETHREAD| |TCEINVALID|
	  |TCENOFILE| |TCENOPERM| |TCEMETA|
	  |TCERHEAD| |TCEOPEN| |TCECLOSE|
	  |TCETRUNC| |TCESYNC| |TCESTAT|
	  |TCESEEK| |TCEREAD| |TCEWRITE|
	  |TCEMMAP| |TCELOCK| |TCEUNLINK|
	  |TCERENAME| |TCEMKDIR| |TCERMDIR|
	  |TCEKEEP| |TCENOREC| |TCEMISC|
          )
  )
(select-module dbm.tokyocabinet)

;; Loads extension
(dynamic-load "tokyocabinet")

(define-class <tcbdb-meta> (<dbm-meta>)
  ())
(define-class <tchdb-meta> (<dbm-meta>)
  ())

(define-class <tcbdb> (<dbm>)
  ((raw-tcbdb :accessor raw-tcbdb-of :initform #f)
   (noblock-lock :init-keyword :noblock-lock :initform #f)
   (nolock :init-keyword :nolock :initform #f))
  :metaclass <tcbdb-meta>)
(define-class <tchdb> (<dbm>)
  ((raw-tchdb :accessor raw-tchdb-of :initform #f)
   (noblock-lock :init-keyword :noblock-lock :initform #f)
   (nolock :init-keyword :nolock :initform #f))
  :metaclass <tchdb-meta>)


;;;
;;; dbm interfaces
;;;

(define-method dbm-open ((self <tcbdb>))
  (next-method)
  self
  (unless (slot-bound? self 'path)
    (error "path must be specified in order to open TokyoCabinet BDB database"))
  (when (raw-tcbdb-of self)
    (error "tcbdb ~S already opened" self))
  (let* ((path   (slot-ref self 'path))
	 (nolock (slot-ref self 'nolock))
	 (noblock-lock (slot-ref self 'noblock-lock))
         (omode  (case (slot-ref self 'rw-mode)
		   ((:read) |BDBOREADER|)
		   ((:write) (logior |BDBOCREAT| |BDBOWRITER|
				    (if nolock |BDBONOLCK| 0)
				    (if noblock-lock |BDBOLCKNB| 0)))
		   ((:create) (logior |BDBOWRITER| |BDBOTRUNC| |BDBOCREAT|
				     (if nolock |BDBONOLCK| 0)
				     (if noblock-lock |BDBOLCKNB| 0)))))
	 (raw-tcbdb (tcbdb-new)))
    (and (tcbdb-open raw-tcbdb path omode)
	 (begin
	   (slot-set! self 'raw-tcbdb raw-tcbdb)
	   self)))
  )
(define-method dbm-open ((self <tchdb>))
  (next-method)
  self
  (unless (slot-bound? self 'path)
    (error "path must be specified in order to open TokyoCabinet HDB database"))
  (when (raw-tchdb-of self)
    (error "tchdb ~S already opened" self))
  (let* ((path   (slot-ref self 'path))
	 (nolock (slot-ref self 'nolock))
	 (noblock-lock (slot-ref self 'noblock-lock))
         (omode  (case (slot-ref self 'rw-mode)
		   ((:read) |HDBOREADER|)
		   ((:write) (logior |HDBOCREAT| |HDBOWRITER|
				    (if nolock |HDBONOLCK| 0)
				    (if noblock-lock |HDBOLCKNB| 0)))
		   ((:create) (logior |HDBOWRITER| |HDBOTRUNC| |HDBOCREAT|
				     (if nolock |HDBONOLCK| 0)
				     (if noblock-lock |HDBOLCKNB| 0)))))
	 (raw-tchdb (tchdb-new)))
    (and (tchdb-open raw-tchdb path omode)
	 (begin
	   (slot-set! self 'raw-tchdb raw-tchdb)
	   self)))
  )

(define-method dbm-close ((self <tcbdb>))
  (let ((raw-tcbdb (raw-tcbdb-of self)))
    (and raw-tcbdb (tcbdb-close raw-tcbdb)
	 (begin (slot-set! self 'raw-tcbdb #f)
		#t))))
(define-method dbm-close ((self <tchdb>))
  (let ((raw-tchdb (raw-tchdb-of self)))
    (and raw-tchdb (tchdb-close raw-tchdb)
	 (begin (slot-set! self 'raw-tchdb #f)
		#t))))

(define-method dbm-closed? ((self <tcbdb>))
  (not (slot-ref self 'raw-tcbdb)))
(define-method dbm-closed? ((self <tchdb>))
  (not (slot-ref self 'raw-tchdb)))

(define-method dbm-put! ((self <tcbdb>) key value)
  (next-method)
  (let ((key (%dbm-k2s self key))
	(value (%dbm-v2s self value)))
    (cond [(tcbdb-put (raw-tcbdb-of self) key value)
           (undefined)]
          [else
           (error (if (eq? :read (slot-ref self 'rw-mode))
                    "tokyocabinet bdb: database is opened as read only"
                    "tcbdb-put failed"))])))

(define-method dbm-put! ((self <tchdb>) key value)
  (next-method)
  (let ((key (%dbm-k2s self key))
	(value (%dbm-v2s self value)))
    (let1 ret (tchdb-put2 (raw-tchdb-of self) key value)
    (if ret
	ret
	(if (eq? :read (slot-ref self 'rw-mode))
	    (error "tokyocabinet hdb: database is opened as read only")
	    #f)))))

(define-method dbm-get ((self <tcbdb>) key . args)
  (next-method)
  (let ((key (%dbm-k2s self key)))
    (let1 ret (tcbdb-get (raw-tcbdb-of self) key)
      (if ret
	  (%dbm-s2v self ret)
	  (if (null? args)
	      (errorf "tokyocabinet bdb: not data for key ~s in database ~s"
		      key self)
	      (car args))))))
(define-method dbm-get ((self <tchdb>) key . args)
  (next-method)
  (let ((key (%dbm-k2s self key)))
    (let1 ret (tchdb-get2 (raw-tchdb-of self) key)
      (if ret
	  (%dbm-s2v self ret)
	  (if (null? args)
	      (errorf "tokyocabinet hdb: not data for key ~s in database ~s"
		      key self)
	      (car args))))))

(define-method dbm-exists? ((self <tcbdb>) key)
  (and (tcbdb-get (raw-tcbdb-of self) (%dbm-k2s self key)) #t))
(define-method dbm-exists? ((self <tchdb>) key)
  (and (tchdb-get2 (raw-tchdb-of self) (%dbm-k2s self key)) #t))

(define-method dbm-delete! ((self <tcbdb>) key)
  (or (tcbdb-out2 (raw-tcbdb-of self) (%dbm-k2s self key))
      (if (eq? :read (slot-ref self 'rw-mode))
	  (error "tokyocabinet bdb: database is opened as read only")
	  #f)))
(define-method dbm-delete! ((self <tchdb>) key)
  (or (tchdb-out2 (raw-tchdb-of self) (%dbm-k2s self key))
      (if (eq? :read (slot-ref self 'rw-mode))
	  (error "tokyocabinet hdb: database is opened as read only")
	  #f)))

(define-method dbm-fold ((self <tcbdb>) proc knil)
  (let* ((raw-tcbdb (raw-tcbdb-of self))
	 (cur (tcbdbcur-new raw-tcbdb)))
    (if (not (tcbdbcur-first cur))
	knil
	(let loop((continue? #t) (ret knil))
	  (if (not continue?)
	      ret
	      (let1 newret (proc (%dbm-s2k self (tcbdbcur-key cur))
				 (%dbm-s2v self (tcbdbcur-val cur))
				 ret)
		(loop (tcbdbcur-next cur) newret)))))))
(define-method dbm-fold ((self <tchdb>) proc knil)
  (let1 raw-hdb (raw-tchdb-of self)
    (unless (tchdb-iterinit raw-hdb)
      (error "tokyocabint hdb: iter init failed!"))
    (let loop((continue? #t) (ret knil))
      (if (not continue?)
	  ret
	  (receive (cont? key val)
	      (tchdb-iternext3 raw-hdb)
	    (loop cont?
		  (proc (%dbm-s2k self key)
			(%dbm-s2v self val)
			ret)))))))


;; Epilogue
(provide "dbm/tokyocabinet")

